.welcome, .score {
  font-size:2em;
}

.high_score, .score {
  padding: 0.2em;
  font-size: 2.2em;
}

.score {
  border: 1px solid shade(grey, 0.5);
  border-top: 0px;
  border-bottom-left-radius: 12px;
  border-bottom-right-radius: 12px;
}

.gameover {
  font-size:3em;
}

.goal {
  font-size:1.2em;
}

.tip {
  font-size: 1em;
}
 
.play {
  font-size: 1.5em;
  padding-right:12px;
  padding-left:12px;
  padding-top:6px;
  padding-bottom:6px;
}
